import httpx

from meta_paper.adapters._base import PaperListing, PaperDetails, PaperMetadataAdapter
from meta_paper.search import QueryParameters


class SemanticScholarAdapter(PaperMetadataAdapter):
    BASE_URL = "https://api.semanticscholar.org/graph/v1"

    def __init__(self, http_client: httpx.AsyncClient, api_key: str | None = None):
        self.__http = http_client
        self.__request_headers = {} if not api_key else {"x-api-key": api_key}

    @property
    def request_headers(self) -> dict:
        return self.__request_headers

    async def search(self, query: QueryParameters) -> list[PaperListing]:
        search_endpoint = f"{self.BASE_URL}/paper/search"
        query_params = query.semantic_scholar().set(
            "fields", "title,externalIds,authors"
        )
        response = await self.__http.get(
            search_endpoint, headers=self.__request_headers, params=query_params
        )
        response.raise_for_status()

        search_results = response.json().get("data", [])
        return [
            PaperListing(
                doi=paper_info["externalIds"]["DOI"],
                title=paper_info["title"],
                authors=author_names,
            )
            for paper_info in search_results
            if self.__has_valid_doi(paper_info)
            and paper_info.get("title")
            and (author_names := self.__get_author_names(paper_info))
        ]

    async def details(self, doi: str) -> PaperDetails:
        doi = self.__prepend_doi_prefix(doi)
        paper_details_endpoint = f"{self.BASE_URL}/paper/{doi}"
        params = {"fields": "title,authors,references,abstract,externalIds"}
        response = await self.__http.get(
            paper_details_endpoint, headers=self.__request_headers, params=params
        )
        response.raise_for_status()

        paper_data = response.json()
        if not (title := paper_data.get("title")):
            raise ValueError("paper title missing")
        if not (authors := self.__get_author_names(paper_data)):
            raise ValueError("paper authors missing")
        if not (abstract := paper_data.get("abstract")):
            raise ValueError("paper abstract missing")

        return PaperDetails(
            doi=doi,
            title=title,
            authors=authors,
            abstract=abstract,
            references=[
                self.__prepend_doi_prefix(ref["externalIds"]["DOI"])
                for ref in paper_data["references"]
                if self.__has_valid_doi(ref)
            ],
        )

    @staticmethod
    def __prepend_doi_prefix(doi: str) -> str:
        doi = doi.strip()
        if not doi.startswith("DOI:"):
            doi = f"DOI:{doi}"
        return doi

    @staticmethod
    def __has_valid_doi(paper_info: dict) -> bool:
        return (
            "externalIds" in paper_info
            and "DOI" in paper_info["externalIds"]
            and paper_info["externalIds"]["DOI"]
        )

    @staticmethod
    def __get_author_names(author_data: dict) -> list[str]:
        if not author_data.get("authors"):
            return []
        if not any("name" in x and x["name"] for x in author_data["authors"]):
            return []

        return list(map(str, (author["name"] for author in author_data["authors"])))



from dataclasses import dataclass
from typing import Protocol

from meta_paper.search import QueryParameters


@dataclass
class PaperListing:
    doi: str
    title: str
    authors: list[str]


@dataclass
class PaperDetails:
    doi: str
    title: str
    authors: list[str]
    abstract: str
    references: list[str]


class PaperMetadataAdapter(Protocol):
    async def search(self, query: QueryParameters) -> list[PaperListing]:
        pass

    async def details(self, doi: str) -> PaperDetails:
        pass



import asyncio
import itertools
from collections.abc import Sequence
from typing import Iterable, Generator

import httpx

from meta_paper.adapters import (
    OpenCitationsAdapter,
    SemanticScholarAdapter,
    PaperListing,
    PaperDetails,
    PaperMetadataAdapter,
)
from meta_paper.search import QueryParameters


class PaperMetadataClient:
    def __init__(self, http_client: httpx.AsyncClient | None = None):
        self.__providers: list[PaperMetadataAdapter] = []
        self.__http = http_client or httpx.AsyncClient(headers={
            "Accept": "application/json",
            "Accept-Encoding": "deflate,gzip;q=1.0",
        })

    @property
    def providers(self) -> Sequence[PaperMetadataAdapter]:
        return self.__providers

    def use_open_citations(self, token: str | None = None) -> "PaperMetadataClient":
        """Add OpenCitations adapter to the client."""
        self.__providers.append(OpenCitationsAdapter(self.__http, token))
        return self

    def use_semantic_scholar(self, api_key: str | None = None):
        """Add SemanticScholar adapter to the client."""
        self.__providers.append(SemanticScholarAdapter(self.__http, api_key))
        return self

    async def search(self, query: QueryParameters) -> list[PaperListing]:
        """Perform an asynchronous search across all providers."""
        tasks = [provider.search(query) for provider in self.providers]
        results = await asyncio.gather(*tasks)
        results = list(itertools.chain.from_iterable(results))
        return list(self.__dedupe_by_doi(results))

    async def details(self, doi: str) -> PaperDetails:
        """Fetch paper summaries asynchronously from all providers."""
        tasks = [provider.details(doi) for provider in self.providers]
        summaries = await asyncio.gather(*tasks)
        details = {}
        refs = set()
        authors = set()
        for summary in summaries:
            details["doi"] = summary.doi or details.get("doi", "")
            details["title"] = summary.title or details.get("title", "")
            details["abstract"] = summary.abstract or details.get("abstract", "")
            for ref in summary.references:
                refs.add(ref)
            for author in summary.authors:
                if author:
                    authors.add(author)
        return PaperDetails(**details, references=list(refs), authors=list(authors))

    @staticmethod
    def __dedupe_by_doi(
        results: Iterable[PaperListing],
    ) -> Generator[PaperListing, None, None]:
        """Remove duplicates based on DOI or title."""
        seen = set()
        for result in results:
            if result.doi in seen:
                continue
            seen.add(result.doi)
            yield result



import re

import httpx

from meta_paper.adapters._base import PaperDetails, PaperListing, PaperMetadataAdapter
from meta_paper.search import QueryParameters


class OpenCitationsAdapter(PaperMetadataAdapter):
    REFERENCES_REST_API = "https://opencitations.net/index/api/v2"
    META_REST_API = "https://w3id.org/oc/meta/api/v1"
    DOI_RE = re.compile(r"\b(doi:[0-9a-z./]+)\b", re.IGNORECASE)

    def __init__(
        self, http_client: httpx.AsyncClient, api_token: str | None = None
    ) -> None:
        self.__http = http_client
        self.__headers = {} if not api_token else {"Authorization": api_token}

    @property
    def http_headers(self):
        return self.__headers

    async def search(self, _: QueryParameters) -> list[PaperListing]:
        return []

    async def details(self, doi: str) -> PaperDetails:
        """Fetch references and citations for a DOI."""

        doi = doi.strip()
        if not doi.startswith("doi:"):
            doi = f"doi:{doi}"

        response = await self.__http.get(
            f"{self.REFERENCES_REST_API}/references/{doi}", headers=self.__headers
        )
        response.raise_for_status()

        refs = [
            self.DOI_RE.search(ref["cited"]).group(1)
            for ref in response.json()
            if self.DOI_RE.search(ref["cited"])
        ]

        response = await self.__http.get(
            f"{self.META_REST_API}/metadata/{doi}", headers=self.__headers
        )
        response.raise_for_status()
        metadata = next(iter(response.json()))

        return PaperDetails(
            doi=doi,
            title=metadata["title"],
            authors=metadata.get("authors", "").split(";"),
            abstract="",
            references=refs,
        )
